/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <time.h>
#include <getopt.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
//#include <sys/wait.h>
#include <sys/stat.h>
#include <errno.h>

#include <curses.h>
#include <readline/readline.h>
#include <readline/history.h>

#include "my_stdio.h"
#include "my_math.h"
#include "camera.h"
#include "fits_io.h"
#include "version.h"
#include "camera_tools.h"
#include "camera_xpa.h"

int cam;
int done = 0;

int readline_flag = 1;
int readline_history_flag = 1;
int interactive_flag = 1;
int auto_win_flag = 1;
int toggle_pal = 1;

const char *name_definition = {"Star I CCD Camera acquisition system"};
static const char *help_text[] ={
#       ifdef VERBOSE_MODE
        "  -v\t\t - verbose mode (incremental).",
#       endif
        "  -n [--interac] - Not interactive mode.",
        "  -r\t\t - Don't use \"read line\" history.",
	"  -V\t\t - Show program version.",
	"  --link\t - Make symbolic links for each command.",
	"  --unlink\t - Remove symbolic links for commands.",
        "  -h [--help]\t - this help !!!.",
        ""
};

/* The names of functions that actually do the manipulation. */
int com_help  __P((char *));
int com_quit  __P((char *));
int com_test  __P((char *));
int com_time  __P((char *));
int com_temperature  __P((char *));
int com_etime  __P((char *));
int com_exposure  __P((char *));
int com_flat  __P((char *));
int com_palette  __P((char *));
int com_get  __P((char *));
int com_focus  __P((char *));
int com_shade  __P((char *));
int com_clear  __P((char *));
int com_gain  __P((char *));
int com_dark  __P((char *));
int com_bias  __P((char *));
int com_show_pos __P((char *));
int com_config __P((char *));
int com_display __P((char *));
int com_note __P((char *));

int com_temp_reg __P((char *));

int com_nroi __P((char *));
int com_roi_expose __P((char *));
int com_roi __P((char *));

int com_box_x0 __P((char *));
int com_box_y0 __P((char *));
int com_box_x_len __P((char *));
int com_box_y_len __P((char *));

int com_box_get __P((char *));
int com_box_set __P((char *));

int com_mouse_get __P((char *));
int com_chdir __P((char *));
int com_export __P((char *));


/* A structure which contains information on the commands this program
   can understand. */

typedef struct {
	char *name;		/* User printable name of the function. */
	int link;               /* Make link for this command. */
	int show;               /* Show or hide command. */
	rl_icpfunc_t *func;	/* Function to call to do the job. */
	char *param;            /* Parameters for this function.  */
	char *doc;		/* Documentation for this function.  */
} COMMAND;

COMMAND commands[] = {
	{ "test",     0, 0, com_test,"", "[!]" },
	{ "time",     0, 1, com_time,"", "Set date & time on camera controller" },
	{ "temp",     1, 1, com_temperature,"", "Get current CCD temperature" },
	{ "etime",    1, 1, com_etime, "[time]", "Get/Set exposure time (sec)" },
	{ "treg",     1, 1, com_temp_reg,"[on|off]", "Get/Set teperature regulator" },
	{ "expose",   1, 1, com_exposure,"[etime]", "Exposure (F5)" },
	{ "dark",     1, 1, com_dark,"[etime]", "Dark integration" },
	{ "bias",     1, 1, com_bias,"", "Bias integration" },
	{ "flat",     1, 1, com_flat,"[etime]", "Flat integration" },

	{ "get",      1, 1, com_get,"{filename[."IMAGE_EXTENSION"]}", "Get current image from the camera controller" },
	{ "display",  0, 1, com_display,"{filename[."IMAGE_EXTENSION"]}", "Display filename."IMAGE_EXTENSION" image" },
	{ "note",     1, 1, com_note,"{filename[."IMAGE_EXTENSION"] note text}", "Put note in header of filename."IMAGE_EXTENSION },
	{ "focus",    0, 1, com_focus,"", "100x100 pixel focus integration" },
	{ "shade",    0, 1, com_shade,"", "Test video display memory" },
	{ "clear",    0, 1, com_clear,"", "Clear video display memory" },
	{ "gain",     1, 1, com_gain,"[1|4]", "Get/Set system gain factor" },
	{ "showp",    1, 1, com_show_pos,"[&]", "Show current position (HA, Dist. Pol.). &: run in xterm" },
	{ "conf",     0, 1, com_config,"", "Edit current configuration file" },
	{ "pal",      1, 1, com_palette,"[sat|hsat] [[min] max]", "Chage current video palette (none:reset, sat: saturation, hsat: hight saturation; F6|F9: toggle which last palette." },

	{ "nroi",     0, 0, com_nroi,"", "Get/Set ROI number" },
	{ "roiexp",   0, 0, com_roi_expose,"", "Expose ROI" },
	{ "roi",      0, 0, com_roi, "[x0 y0 xlen ylen]", "Get/Set ROI" },

	{ "gpointer", 0, 0, com_mouse_get, "", "Get current pointer location and pixel value"},
	{ "x0",       0, 0, com_box_x0,"", "Get/Set X origin of box" },
	{ "y0",       0, 0, com_box_y0,"", "Get/Set Y origin of box" },
	{ "xlen",     0, 0, com_box_x_len,"", "Get/Set length X of box" },
	{ "ylen",     0, 0, com_box_y_len,"", "Get/Set length Y of box" },

	{ "gbox",     0, 0, com_box_get,"", "Get box" },
	{ "sbox",     0, 0, com_box_set,"", "Set box" },

	{ "cd",       0, 0, com_chdir, "[path]", "Change working directory" },	
	{ "export",   0, 0, com_export, "name=value", "Change or add an environment variable" },	
	{ "help",     0, 0, com_help, "[cmd]", "Display this text" },
	{ "?",        0, 1, com_help, "[cmd]", "Synonym for `help'" },
	{ "quit",     0, 0, com_quit, "", "Quit ..." },
	{ "q",        0, 1, com_quit, "", "Quit :-)" },
	{ (char *)NULL, 0, 0, (rl_icpfunc_t *)NULL, (char *) NULL, (char *)NULL }
};



/* Generator function for command completion.  STATE lets us
   know whether to start from scratch; without any state
   (i.e. STATE == 0), then we start at the top of the list. */
char *command_generator (char *text, int state)
{ 
	static int list_index, len;
	char *name;

	/* If this is a new word to complete, initialize now.  This
	   includes saving the length of TEXT for efficiency, and
	   initializing the index variable to 0. */
	if (!state) {
		list_index = 0;
		len = strlen (text);
	}

	/* Return the next name which partially matches from the
	   command list. */
	while ((name = commands[list_index].name)) {
		list_index++;
		if (strncmp (name, text, len) == 0)
			return (strdup(name));
	}

	/* If no names matched, then return NULL. */
	return ((char *)NULL);
}


char **fileman_completion (const char *text, int start, int end)
{
	char **matches;

	matches = (char **)NULL;

	/* If this word is at the start of the line, then it is a
	   command to complete.  Otherwise it is the name of a file in
	   the current directory.  */
	if (start == 0)
	  matches = rl_completion_matches (text, 
					   (rl_compentry_func_t *) command_generator);
	
	return (matches);
}


/* Tell the GNU Readline library how to complete.  We want to try to
   complete on command names if this is the first word in the line, or
   on filenames if not. */
void initialize_readline (void)
{
	Keymap km;
	/* char f9_macro[]={"\\C-p\\n"}; */ /* last command */ 
	char f9_macro[]={"pal last\\n"};
	char f5_macro[]={"expose\\n"};
	char f6_macro[]={"pal last\\n"};

	/* Allow conditional parsing of the ~/.inputrc file. */
	rl_readline_name = "stari";
	
	/* Tell the completer that we want a crack first. */
	rl_attempted_completion_function = fileman_completion;

	/* Read keybindings and variable assignments from filename */
	/* int rl_read_init_file (const char *filename); */

	km = rl_get_keymap_by_name("emacs"); 
	rl_macro_bind("\e[15~", f5_macro, km);
	rl_macro_bind("\e[17~", f6_macro, km);
	rl_macro_bind("\e[20~", f9_macro, km);
}


int un_link_commands(void)
{
	int i;

	for (i = 0; commands[i].name; i++) {
		if ( commands[i].link ) {
			unlink(commands[i].name);
		}
	}

	return 0;
}


int link_commands_to(char *program_name)
{
	int i;

	for (i = 0; commands[i].name; i++) {
		if ( commands[i].link ) {
			unlink(commands[i].name);
			if (symlink(program_name, commands[i].name) != 0)
				perror("stari: ");
		}
	}

	return 0;
}


/* Look up NAME as the name of a command, and return a pointer to that
   command.  Return a NULL pointer if NAME isn't a command name. */
COMMAND *find_command (char *name)
{
	int i;

	for (i = 0; commands[i].name; i++)
		if (strcmp (name, commands[i].name) == 0)
			return (&commands[i]);

	return ((COMMAND *)NULL);
}


/* Execute a command line. */
int execute_line (char *line)
{
  int i;
	COMMAND *command;
	char *word;
	int pos_null = 0;
	/* Isolate the command word. */
	i = 0;
	while (line[i] && WHITESPACE(line[i]))	i++;
	word = line + i;

	while (line[i] && !WHITESPACE (line[i])) i++;

	if (line[i]){
		pos_null = i;
		line[i++] = '\0';
	}
	command = find_command (word);
	if (!command) {
		if(pos_null)
			line[pos_null] = ' ';
		my_system(line);
		//PVERB(PERR, "%s: No such command for stari.\n", word);
		return (-1);
	}

	/* Get argument to command, if any. */
	while (WHITESPACE (line[i])) i++;

	word = line + i;

	/* Call the function. */
	return ((*(command->func)) (word));
}


/* The user wishes to quit using this program.  Just set DONE
   non-zero. */
int com_quit (char *arg)
{
	done = 1;
	return (0);
}

/* Print out help for ARG, or for all of the commands if ARG is not
   present. */
int com_help (char *arg)
{
	int i, j;
	int printed = 0;
	int cmdlen;
	const int param_legend = 22;

	for (i = 0; commands[i].name; i++) {
		if ((!*arg && commands[i].show) || 
		    (strcmp (arg, commands[i].name) == 0) ) {
			cmdlen = strlen(commands[i].name) + 
				strlen(commands[i].param) + 1;
			printf ("%s %s", commands[i].name, 
				commands[i].param);
			if (cmdlen >= param_legend) {
				printf("\n");
				cmdlen = 0;
			}
			for(j = cmdlen; j < param_legend; j++)
				printf(" ");
			printf ("%s.\n", commands[i].doc);

			printed++;
		}
	}

	if (!printed) {
		printf ("No commands match `%s'.  Possibilties are:\n", arg);

		for (i = 0; commands[i].name; i++) {
			/* Print in six columns. */
			if (printed == 6) {
				printed = 0;
				printf ("\n");
			}
			
			printf ("%s\t", commands[i].name);
			printed++;
		}

		if (printed)
			printf ("\n");
	}
	return (0);
}


#define COMMAND_LINE_LEGEND 80


char *local_readline(char *prompt)
{
        char *line = NULL;
        fd_set rfds;
        struct timeval tv;
        int retval;
        char c;
        int l;

	write(0, prompt, strlen(prompt));
        line = (char *) malloc(COMMAND_LINE_LEGEND); 

        /* Loop reading and executing lines until the user quits. */
        l = 0;
        do {
                FD_ZERO(&rfds);
                FD_SET(0, &rfds);
                tv.tv_sec = 10;
                tv.tv_usec = 0;
                retval = select(50, &rfds, NULL, NULL, &tv);
                c = 0;
                if (retval) {
                        if (FD_ISSET(0, &rfds)) {
                                /* Win telnet clients send char by char ... */
                                if( read(0, &c, 1) != 1) {
                                        free(line);
                                        return NULL;
                                }
                                *(line+l) = c;
                                l++;
                        }
                }
        } while(c != '\n' && l < COMMAND_LINE_LEGEND && !done);
        if ( *(line+l-2) == 0xd ) *(line+l-2) = 0;
        else *(line+l-1) = 0;

        return line;
}


void local_add_history (char *line)
{
	HIST_ENTRY *history;

	if ( history_length ) {
		history = history_get(history_length);
		if (strcmp(line, history->line) == 0) {
			return;
		}
	}
	if (strncmp(line, "q", 1) == 0) {
		return;
	}
	if (strncmp(line, "pal last", 8) == 0) {
		return;
	}
	if (strncmp(line, "q", 1) == 0) {
		return;
	}

	add_history (line);

	return;
}



int command_main(void)
{
	char *line, *s;
	char *history_file = (char *) malloc(_POSIX_PATH_MAX * sizeof(char));
	int err;

	if ( readline_flag ) {
		initialize_readline ();	/* Bind our completer. */
	}
	if ( readline_history_flag ) {
		sprintf(history_file,"%s/%s", getenv("HOME"), USER_HISTORY);
		if (((err = read_history(history_file)) != 0) && (err != 2)) {
			PVERB(PERR, 
				"Error [%d] reading READLINE history file.\n",
				err);
		}
	}
	/* Loop reading and executing lines until the user quits. */
	for ( ; done == 0; ) {
		if ( readline_flag )
			line = readline ("stari> ");
		else 
                        line = local_readline ("stari> ");

		if (!line)
			break;
		s = stripwhite (line);
		if (*s)	{
			if ( readline_flag )
				local_add_history (s);
			execute_line (s);
		}
		free (line);
	}
	if ( readline_history_flag ) {
		if (((err = write_history(history_file)) != 0) && (err != 2)) {
			PVERB(PERR, 
				"Error [%d] writing READLINE history file.\n",
				err);
		}
	}
	free(history_file);

	return (0);
}


int return_code(int ret)
{
	if(ret < 0) {
		if (ret == -2)	printf(" camera busy\n");
		else printf(" fail\n");
	}/*else
		printf(" ret: %d\n", ret);
	 */
	return ret;
}


int show_mini_setup(void)
{
	int gain;
	int etime;

	if (! interactive_flag) return 0;

	cc_get_gain(cam, &gain);
	if(gain == 0) gain = 1;
	else if(gain == 1) gain = 4;
	cc_get_exposure_time(cam, &etime);
	printf("Gain: %d, ETime: %5.1f\n", gain, (double) etime / 10.0);

	return 0;
}


int show_min_max_value(void)
{ 
	int min, max;
	int autow;

	if (! interactive_flag) return 0;

	cc_get_auto_win(cam, &autow);
	if( autow == 1 ) {
		cc_get_win_max(cam, &max);
		cc_get_win_min(cam, &min);
		printf("Palette Min: %4d, Max %4d\n", min, max);
		toggle_pal = 1;
	}
	return 0;
}

int show_warn_tem(int cam)
{
	int t;

	if (warn_temp_flag ) {
		cc_get_temperature(cam, &t);
		if ( t > warn_temp) {
			printf("\n*** WARNING: ");
			printf("Current CCD temperature is hight (%d C)", t);
			printf(" ***%c\n\n", 0x07);

			return -1;
		}
	}

	return 0;
}

int show_config(int imtype)
{
	int ret;
	struct image_2d im;
	//struct fit_tag *ftp;

	if (! interactive_flag) return 0;
	im.ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	im.ft->next = NULL;

	if ((ret = ct_read_header_file(&im, local_config_file)) == 0) {
		printf("Current configuration:\n");
		printf(" OBJECT: %s", 
		       im.object);
		if (imtype)
			printf(", IMAGE TYPE: %s\n", 
			       str_desc[imtype].imagetyp);
		else
			printf("\n");
		printf(" RA: %s, DEC: %s, EQUINOX: %6.1f\n", 
		       im.ra_cat, im.dec_cat, im.equinox);
		printf(" FILTER: %s\n", im.filter);
		printf(" OBSERVER: %s\n", im.observer);
		/*
		for(ftp = im.ft; ftp->next != NULL; ftp = ftp->next) {
			if ( ftp->tag[0] != 0)
				printf(" %s: %s\n", ftp->tag, ftp->str);
		}
		*/
	}
	fits_tag_free(im.ft);

	return ret;
}


int clear_buff(int fd)
{
	fd_set rfds;
        int fd_max = fd + 1;
        int retval;
	struct timeval tv;
	char c;

	do {
		tv.tv_sec = 0;
		tv.tv_usec = 100000;
		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);
		retval = select(fd_max, &rfds, NULL, NULL, &tv);
		if (retval) {
			if( FD_ISSET(fd, &rfds) ) {
				read(fd, &c, 1);
			}
		}
	}while(retval > 0);

	return 0;
}

int key_in_buff(int fd)
{
	fd_set rfds;
        int fd_max = fd + 1;
        int retval;
	struct timeval tv;

	tv.tv_sec = 0;
	tv.tv_usec = 100000;
	FD_ZERO(&rfds);
	FD_SET(fd, &rfds);
	retval = select(fd_max, &rfds, NULL, NULL, &tv);
	if (retval) {
		if( FD_ISSET(fd, &rfds) ) {
			return 1;
		}
	}

	return 0;
}


int show_remaining_time(int time)
{
	int time_sec;
	double t;
	char key;
	struct timeval exp, now;

	time_sec = (int) rint((double) time / 10.0);
	if (interactive_flag) {
		gettimeofday(&now, NULL);
		exp.tv_sec = now.tv_sec + time_sec;
		exp.tv_usec = now.tv_usec + 
			(time_t) (time - time_sec * 10) * 100000;
		while(timercmp(&exp, &now, >)) {
			t = (double) ( (double) (exp.tv_sec - now.tv_sec) +
				       (double) (exp.tv_usec - now.tv_usec) / 
				       1000000.0);
			printf("remaining %5.1f sec. [ESC ENTER] to abort  \r",
			       t);
			fflush(stdout);
			if (key_in_buff(1)) {
				read(1, &key, 1);
				//printf("Key: 0x%x\n", key);
				if( key == 0x1b ) {
					printf("aborting...\n");
					fflush(stdout);
					return 1;
				}
			}
			gettimeofday(&now, NULL);
		}
		printf("                                                      \r");
		fflush(stdout);
	} else {
		sleep(time_sec);
	}

	return 0;
}




int com_time (char *arg)
{
	time_t now;

	time(&now);
	cc_set_date_time(cam, now);

	return 0;
}


int com_temperature (char *arg)
{ 
 	int t;
	int ret;

	ret = cc_get_temperature(cam, &t);
	if (ret == 0) {
		if (interactive_flag)
			printf(" Currente camera head temperature is ");
		printf("%d\n", t);
	}
	else return_code(ret);

	return 0;
}


int com_etime (char *arg)
{
	int etime;
	int ret = -1;

        if (arg[0] == 0) {
		ret = cc_get_exposure_time(cam, &etime);
		if(ret == 0) {
			if (interactive_flag)
				printf(" Current exposure time is ");
			printf("%.1f\n", (double) etime / 10.0);
		} else return_code(ret);
                return ret;
        }
	if (strisnum(arg) > 0) {
		etime = (int) my_round(atof(arg) * 10.0);
		if ( etime > 0)
			cc_set_exposure_time(cam, etime);
	}else{
		PVERB(PERR, " etime: bad argument %s\n", arg);
		ret = -1;
	}
	return ret;
}


int com_gain (char *arg)
{
	int gain;

        if (arg[0] == 0) {
		cc_get_gain(cam, &gain);
		if ( gain == GAIN_HIGH ) gain = 4;
		else gain = 1;
		if (interactive_flag)
			printf(" Current gain is ");
                printf("%d\n", gain);
                return 0;
        }

	gain = atoi(arg);
	switch(gain){
	case 1:
		gain = GAIN_LOW;
		break;
	case 4:
		gain = GAIN_HIGH;
		break;
	default:
		PVERB(PERR, " gain: bad argument %s\n", arg);
		return -1;
	}

	return return_code( cc_set_gain(cam, gain) );
}


int com_temp_reg (char *arg)
{
	int reg;
	char sta[4];

        if (arg[0] == 0) {
		cc_get_temp_regulation(cam, &reg);
		if ( reg == TEMP_REG_ON ) sprintf(sta,"on");
		else sprintf(sta,"off");
		if (interactive_flag)
			printf(" temperature regulator is ");
                printf("%s\n", sta);
                return 0;
        }

	if( strncmp(arg, "on", 2) == 0)
		reg = TEMP_REG_ON;
	else if(strncmp(arg, "off", 3) == 0) 
		reg = TEMP_REG_OFF;
	else {
		PVERB(PERR, " tempreg: bad argument %s\n", arg);
		return -1;
	}
	cc_set_temp_regulation(cam, reg);

	return 0;
}



int com_exposure (char *arg)
{
	int etime;

	if (arg[0] != 0) {
		if (strisnum(arg) > 0) {
			com_etime(arg);
		} else {
			PVERB(PERR, " expose: bad argument %s\n", arg);
			return -1;
		}
	}
	show_config(IMT_IMAGE);
	show_warn_tem(cam);
	cc_get_exposure_time(cam, &etime);
	cc_act_exposure(cam);
	if( show_remaining_time(etime) == 1)
		reset_cc(cam);
	else{
		show_mini_setup();
		show_min_max_value();
	}

	return 0;
}


int com_dark (char *arg)
{
	int etime;

	if (arg[0] != 0) {
		if (strisnum(arg) > 0) {
			com_etime(arg);
		} else {
			PVERB(PERR, " dark: bad argument %s\n", arg);
			return -1;
		}
	}
	show_config(IMT_DARK);
	show_warn_tem(cam);
	cc_get_exposure_time(cam, &etime);
	cc_act_dark(cam);
	if( show_remaining_time(etime) == 1)
		reset_cc(cam);
	else{
		show_mini_setup();
		show_min_max_value();
	}

	return 0;
}


int com_bias (char *arg)
{
	show_config(IMT_BIAS);
	show_warn_tem(cam);
	cc_act_bias(cam);
	show_mini_setup();
	show_min_max_value();
	return 0;
}


int com_flat (char *arg)
{
	int etime;

	if (arg[0] != 0) {
		if (strisnum(arg) > 0) {
			com_etime(arg);
		} else {
			PVERB(PERR, " flat: bad argument %s\n", arg);
			return -1;
		}
	}
	show_config(IMT_FLAT);
	show_warn_tem(cam);
	cc_get_exposure_time(cam, &etime);
	cc_act_flat(cam);
	if( show_remaining_time(etime) == 1)
		reset_cc(cam);
	else{
		show_mini_setup();
		show_min_max_value();
	}

	return 0;
}



int com_get (char *arg)
{
	char filename[_POSIX_PATH_MAX];
	time_t t;
	int good_image_flag = 0;
	struct stat dummy;
	char buff[8];
	int i;
	int point;
	int ret = -1;

	if (arg[0] == 0) {
		time(&t);
		sprintf(filename,"im%10ld.%s", t, IMAGE_EXTENSION); 
		
	}else{
		point = 0;
		for(i=0;i<strlen(arg);i++) {
			if(*(arg+i) == '.') {
				point = 1;
				break;
			}
		}
		if(point)
			sprintf(filename,"%s", arg);
		else
			sprintf(filename,"%s.%s", 
				stripwhite(arg), IMAGE_EXTENSION);
	}
	
	if (stat(filename, &dummy) == 0 && interactive_flag) {
		clear_buff(1);
		printf("File %s exist, overwrite (N/y)?: ", filename);
		fflush(stdout);
		read(1, &buff, 1);
		clear_buff(1);
		switch (*buff) {
		case 'Y':
		case 'y':
		case 'S':
		case 's':
			break;
		default:
			printf(" Abort\n");
			return -1;
			break;
		}
	}
	switch (ct_read_save_file(cam, filename)) {
	case CT_IM_OK:
		good_image_flag = 1;
		ret = 0;
		if (interactive_flag)
			printf("ok.\n");
		break;
	case CT_IM_SETUP:
		PVERB(PERR, " Error reading image setup.\n");
		break;
	case CT_IM_DIM:
		PVERB(PERR, " Invalid dimension.\n");
		break;
	case CT_IM_READ:
		PVERB(PERR, " Error reading image.\n");
		break;
	case CT_IM_FITS_ERR:
		PVERB(PERR, " FITS Error.\n");
		break;
	default:
		PVERB(PERR, " Unknown Error.\n");
	}
	/* si llega hasta aca despliego... */
	if (good_image_flag) {
		caxpa_display(filename);
	}
	
	return ret;
}


int com_display (char *arg)
{
	char filename[_POSIX_PATH_MAX];
	int ret = -1;

	if (file_name_exist(filename, arg) > 0) {
		ret = caxpa_display(filename);
	} else {
		printf("File %s[.%s] not found\n", arg, 
		       IMAGE_EXTENSION);
	}

	return ret;
}


int com_note (char *arg)
{ 
	char filename[_POSIX_PATH_MAX];
	struct fit_tag *ft;
	int i;
	char *file;
	char *str;
	char *p;

	i = 0;
	while (arg[i] && WHITESPACE(arg[i])) i++;
	file = arg + i;
	while (arg[i] && !WHITESPACE (arg[i])) i++;
	if (arg[i]){
		arg[i++] = '\0';
	}
	while (arg[i] && WHITESPACE(arg[i])) i++;
	str = arg + i;
	if ( *str == 0) {
		PVERB(PERR, " Bad text.\n");
		return -1;
	}
	while((p = strchr(str, ' ')) != NULL) *p = '_'; 

	ft = (struct fit_tag *) malloc(sizeof(struct fit_tag));
	ft->next = NULL;
	fits_tag_push(ft, "NOTE", str, "");
	if (file_name_exist(filename, file) > 0) {
		image_add_tag(filename, ft);
	} else {
		PVERB(PERR, "File %s[.%s] not found\n", arg, 
		       IMAGE_EXTENSION);
	}
	fits_tag_free(ft);

	return 0;
}


int com_focus (char *arg)
{
	/* FIXME: debe ejecutarlo continuamante hasta que se toque una
           tecla */
	cc_act_focus(cam);
	return 0;
}


int com_shade (char *arg)
{
	cc_act_shade(cam);
	return 0;
}


int com_clear (char *arg)
{
	cc_act_clear(cam);
	return 0;
}


int com_nroi (char *arg)
{
	int v;

        if (arg[0] == 0) {
		cc_get_roi_index(cam, &v);
		if (interactive_flag)
			printf(" Current number of ROIs is ");
		printf("%d\n", v);
                return 0;
        }

	if (strisnum(arg) > 0) {
		v = atoi(arg);
		cc_set_roi_index(cam, v);
	} else {
		PVERB(PERR, " roi: bad argument %s\n", arg);
		return -1;
	}
	
	return 0;
}

int com_roi_expose (char *arg)
{
	int etime;

	if (arg[0] != 0) {
		if (strisnum(arg) > 0) {
			com_etime(arg);
		} else {
			PVERB(PERR, " roiexp: bad argument %s\n", arg);
			return -1;
		}
	}
	cc_get_exposure_time(cam, &etime);
	cc_act_roi_acquisition(cam);
	if( show_remaining_time(etime) == 1)
		reset_cc(cam);
	else{
		show_mini_setup();
		show_min_max_value();
	}
	
	return 0;
}


int com_roi (char *arg)
{
	int xy[4], i;
	char *argp;

        if (arg[0] == 0) {
		cc_get_box_x0(cam, &xy[0]);
		cc_get_box_y0(cam, &xy[1]);
		cc_get_box_x_len(cam, &xy[2]);
		cc_get_box_y_len(cam, &xy[3]);
		if (interactive_flag)
			printf(" Current ROI is "); 
                printf("%d %d %d %d\n", xy[0], xy[1], xy[2], xy[3]);
		
                return 0;
        }

	for(i = 0; i < 4 && arg != NULL; i++) {
		xy[i] = (int) strtod(arg, &argp);
		arg = argp;
	}
	if ( i != 4 ) {
		PVERB(PERR, " roi: bad argument %s\n", arg);
		return -1;
	}
	cc_set_box_x0(cam, xy[0]);
	cc_set_box_y0(cam, xy[1]);
	cc_set_box_x_len(cam, xy[2]);
	cc_set_box_y_len(cam, xy[3]);
	cc_act_set_box(cam);
	return 0;
}


int com_box_x0 (char *arg)
{
	int v;

        if (arg[0] == 0) {
		cc_get_box_x0(cam, &v);
		if (interactive_flag)
			printf(" X origin is ");
                printf("%d\n", v);
                return 0;
        }
	
	if (strisnum(arg) > 0) {
		v = atoi(arg);
		cc_set_box_x0(cam, v);
	} else {
		PVERB(PERR, " x0: bad argument %s\n", arg);
		return -1;
	}

	return 0;
}


int com_box_y0 (char *arg)
{
	int v;

        if (arg[0] == 0) {
		cc_get_box_y0(cam, &v);
		if (interactive_flag)
                printf(" Y origin  is ");
                printf("%d\n", v);
                return 0;
        }

	if (strisnum(arg) > 0) {
		v = atoi(arg);
		cc_set_box_y0(cam, v);
	} else {
		PVERB(PERR, " y0: bad argument %s\n", arg);
		return -1;
	}

	return 0;
}


int com_box_x_len (char *arg)
{
	int v;

        if (arg[0] == 0) {
		cc_get_box_x_len(cam, &v);
		if (interactive_flag)
			printf(" Length of X is ");
                printf("%d\n", v);
                return 0;
        }

	if (strisnum(arg) > 0) {
		v = atoi(arg);
		cc_set_box_x_len(cam, v);
	} else {
		PVERB(PERR, " xlen: bad argument %s\n", arg);
		return -1;
	}

	return 0;
}


int com_box_y_len (char *arg)
{
	int v;

        if (arg[0] == 0) {
		cc_get_box_y_len(cam, &v);
		if (interactive_flag)
			printf(" Length of Y is ");
                printf("%d\n", v);
                return 0;
        }

	if (strisnum(arg) > 0) {
		v = atoi(arg);
		cc_set_box_y_len(cam, v);
	} else {
		PVERB(PERR, " ylen: bad argument %s\n", arg);
		return -1;
	}

	return 0;
}


int com_box_get (char *arg)
{
  cc_act_get_box(cam);

  return 0;
}


int com_box_set (char *arg)
{
  cc_act_set_box(cam);

  return 0;
}


int com_mouse_get (char *arg)
{
	struct mouse_pointer mp;

	cc_act_rcv_mouse_xyz(cam, &mp);
	printf(" x = %3d, y = %3d, val = %4d\n", mp.x, mp.y, mp.val);

	return 0;
}


int com_chdir (char *arg)
{
	char *path;
	int ret;

	if (*arg == 0)
		path = getenv("HOME");
	else
		path = arg;
	ret = chdir(path);
	if( ret != 0)
		perror("cd");

	return ret;
}


int com_export (char *arg)
{
	int ret;
	char *name, *value, *comment;

	if ((ret = parse_line(arg, &name, &value, &comment)) == 0) {
		if ((ret = setenv(name, value,1)) != 0)
			perror("export");
	}
	return ret;
}


int com_palette (char *arg)
{
	int  min, max;
	int  min_fix, max_fix;
	char lut[4096];
	int i;
	int sat_flag = 0;
	char *np = stripwhite(arg);
        char *endp;
	int arg_len;
	int max_set = 0;
	int min_set = 0;

	static char *last_pal = NULL;

	if (np[0] == 0) {
		cc_set_auto_win(cam, 1);
		cc_act_create_lut(cam);
		toggle_pal = 1;
	}else{
		//cc_set_auto_win(cam, 1);
		cc_get_win_min(cam, &min);
		cc_get_win_max(cam, &max);
		if (interactive_flag)
			printf("Camera auto palette: min = %d, max = %d\n",
			       min, max);  
		min_fix = -1;
		max_fix = -1;
		if (strncmp(np, "sat", 3) == 0) {
			sat_flag = 1;
			np += 3;
		} else if (strncmp(np, "hsat", 4) == 0) {
			sat_flag = 2;
			np += 4;
		} else 	if(strncmp(np, "last", 4) == 0) {
			if (interactive_flag)
				printf("Using: pal ");
			if (toggle_pal) {
				if (last_pal != NULL) {
					if (interactive_flag)
						printf("%s\n", last_pal);
					com_palette(last_pal);
				} else {
					char sat[] = "sat";
					if (interactive_flag)
						printf("%s\n", sat);
					com_palette(sat);
				}
			} else {
				if (interactive_flag)
					printf("\n");
				com_palette("");
			}
			//toggle_pal ^= 1;
			return 0;
		}
		arg_len = strlen(arg);
		if ( last_pal != arg ) {
			if ( last_pal != NULL) {
				free(last_pal);
				last_pal = NULL;
			}
			last_pal = (char *) malloc((arg_len+1) * sizeof(char));
			memcpy(last_pal, arg, arg_len);
			*(last_pal+arg_len) = 0;
		}
		if (strlen(np)) {
			min_fix = (int) strtol(np, &endp, 10 );
			min_set = 1;
			if(*endp != 0) {
				np = endp;
				max_fix = (int) strtol(np, &endp, 10 );
				max_set = 1;
			}else{
				max_fix = min_fix;
				max_set = 1;
				min_set = 0;
				min_fix = -1;
			}
		}
		switch(sat_flag) {
		case 0: 
			if (min_fix == -1) min_fix = min;
			if (max_fix == -1) max_fix = max;
			for(i=min_fix;i<max_fix-1;i++)
				lut[i] = (char) 
					rint((double)(i-min_fix) * 255.0/
					     (double)(max_fix - min_fix));
			for(i=1;i<min_fix;i++) lut[i] = 0;
			for(i=max_fix-1;i<4096;i++) lut[i] = 255;
			lut[0] = 128;
			lut[4095] = 255;
			break;
		case 1: /* sat */
			if (min_fix == -1) min_fix = min;
			if (max_fix == -1) max_fix = max;
			//if(min_set)
			for(i=1;i<min_fix;i++) lut[i] = 0;
			for(i=min_fix;i<max_fix-1;i++)
				lut[i] = (char) 
					rint((double)(i-min_fix) * 255.0/
					     (double)(max_fix - min_fix));
			if(max_set) for(i=max_fix-1;i<4096;i++) lut[i] = 0;
			else for(i=max_fix-1;i<4096;i++) lut[i] = 255;
			lut[0] = 255;
			lut[4095] = 0;
			break;
		case 2: /* hsat */
			/* FIXME: */
			//printf("\n\n*** NOt implemented yet ***\n\n");
			//min_fix = 4000;
			//max_fix = 4095;
			for(i=1;i<min_fix;i++) lut[i] = 0;
			for(i=min_fix;i<max_fix-1;i++) lut[i] = 255;
			for(i=max_fix;i<4096;i++) lut[i] = 0;
			lut[0] = 255;
			lut[4095] = 0;
			break;
		}
		if(!auto_win_flag)
			cc_set_auto_win(cam, 0);
		cc_act_send_lut(cam, lut);
		toggle_pal = 0;
	}

	return 0;
}


#if 0
int com_palette_anda (char *arg)
{
	int  min, max;
	char lut[4096];
	int i;
	int pal;
	int hsat = 128;

	if (arg[0] == 0) {
		cc_set_auto_win(cam, 1);
		cc_act_create_lut(cam);
	}else{
		//cc_set_auto_win(cam, 1);
		cc_get_win_min(cam, &min);
		cc_get_win_max(cam, &max);
		printf("Current palette: min = %d, max = %d\n", min, max);  
		if (strncmp(arg, "sat", 3) == 0) {
			for(i=1;i<min;i++) lut[i] = 0;
			for(i=min;i<max-1;i++)
				lut[i] = (char) rint((double)(i-min)*255.0/
						     (double)(max-min));
			for(i=max-1;i<4096;i++) lut[i] = 255;
			lut[0] = 255;
			lut[4095] = 0;
			cc_act_send_lut(cam, lut);
		}else if (strncmp(arg, "hsat", 3) == 0) {
			if(max < hsat) hsat = max; 
			for(i=0;i<max-hsat;i++) lut[i] = hsat;
			for(i=0;i<hsat;i++) lut[max-1-i] = 255-i;
			for(i=max-1;i<4096;i++) lut[i] = 255;
			lut[0] = 255;
			lut[4095] = 0;
			cc_act_send_lut(cam, lut);
		}else{
			pal = atoi(arg);
			if ( pal > 0 && pal <4095) {
				if (min < pal) min = pal;
				if (max < pal) max = pal;
				for(i=min;i<max-1;i++)
					lut[i] = (char) rint((double)(i-min)*255.0/
							     (double)(max-min));
				for(i=1;i<pal;i++) lut[i] = 0;
				for(i=max-1;i<4096;i++) lut[i] = 255;
				lut[0] = 255;
				lut[4095] = 0;
				cc_act_send_lut(cam, lut);
			}
		}

	}

	return 0;
}
#endif

int com_test (char *arg)
{
//	int  min, max;
//	char lut[4096];
//	int i;
#if 0
	int b = atoi(arg);
	char str[20];
	cc_act_move_mouse(cam, b, b);
	
	get_int_from_cc(cam, "W", &b);
	printf("  Inten: %d\n", b);
	put_int_to_cc(cam,"V", 100);
	memset(str, 0, 20);
	
	if( b != 0) return 0;
	
	get_int_from_cc(cam, arg, &b);
	printf("  Result: %d\n", b);
#endif
	//system("vim stari.conf");
	//my_exec("vim");
	cc_set_auto_win(cam, 0);
	cc_set_win_min(cam, 2000);
	cc_set_win_max(cam, 3500);
	cc_act_create_lut(cam);


	return 0;
}


int com_show_pos (char *arg)
{
	char buff[160];
	double ra, dec;

	get_object_position(&ra, &dec);
	if ( *arg == '&') {
		sprintf(buff,"xterm -e %s -l %f -r %f -d %f&",
			TRACK_COMMAND, geo_loc.longitude, ra, dec);

	}else{
		sprintf(buff,"%s -l %f -r %f -d %f",
			TRACK_COMMAND, geo_loc.longitude, ra, dec);
	}
	return system(buff);
}


int com_config (char *arg)
{
	int ret;
	char buff[160];
	char *editor = getenv("EDITOR");
	
	if( editor == NULL) editor = DEF_EDITOR;
	sprintf(buff,"%s  %s", editor, local_config_file);

	ret = system(buff);
	show_config(0);
	
	return ret;
}


static void cartel(void)
{
	static const char *msg[] =
	{
	"",
	"Camera Control for STAR I",
	"Version: "VERSION,
	"http://www.iar.unlp.edu.ar/~fede/pub/star_i",
	"",
	"(c) Copyright 2004, 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,",
	"Instituto Argentino de Radio Astronomia (IAR).",
	"",
	"This program is free software; you can redistribute it and/or",
	"modify it under the terms of the GNU General Public License as",
	"published by the Free Software Foundation; either version 2 of",
	"the License, or (at your option) any later version.",
	"",     
	"The author does NOT admit liability nor provide warranty for any",
	"of this software. This material is provided \"AS-IS\" in the hope",
	"that it may be useful for others.",
	""
	};
	unsigned j;
	for (j = 0; j < sizeof(msg)/sizeof(msg[0]); j++)
		fprintf(stderr, "\t%s\n", msg[j]);
	
    return;
}


void version(char *program_name)
{
        fprintf(stderr,"%s: ", name_definition);
        fprintf(stderr,"%s version %s\n", program_name, VERSION);
}                       


void usage(char *program_name)
{         
        version(program_name);
        fprintf(stderr,
#               ifdef VERBOSE_MODE
 "Usage: \t%s [-vnrhV][--interac][--help]\n",
#                else
 "Usage: \t%s [-nrhV][--interac][--help]\n",
#               endif   
                program_name);
}


void help()
{       
        int i=0;
        
        while((char) *help_text[i] != 0)
                fprintf(stderr,"%s\n",help_text[i++]);
        fprintf(stderr,"\n");
}




int main(int argc, char * const argv[])
{
	char line[_POSIX_ARG_MAX];
	COMMAND *command;
	int ret = -1;
	int op;
	char *autowin = getenv("AUTOWIN");
	char *program_name;
	char *cp;
	struct option const longopts[] =
        {
                {"help", 0, 0, 'h'},
                {"verbose", 0, 0, 'v'},
                {"interac", 0, 0, 'n'},
		{"link", 0, 0, 201},
		{"unlink", 0, 0, 202},
                {0,  0,  0,  0}
        };

	/* who is my name? */
        if ((cp = strrchr(argv[0], '/')) != NULL) program_name  = cp + 1;
        else program_name  = argv[0];

	verbose_flag = PERR;
	while  ((op = getopt_long(argc, argv, "vnrhV",longopts, 0)) != EOF)
		switch( op ){
		case 'n':
                        readline_flag = 0;
			readline_history_flag = 0;
			interactive_flag = 0;
                        break;
		case 'r':
			readline_history_flag = 0;
                        break;
		case 'v':
			verbose_flag = (verbose_flag << 1) | 1;
			break;
		case 'V':
                        version(program_name);
                        exit(-1);
                        break;
		case 201:
			link_commands_to(argv[0]/*program_name*/);
			exit(0);
		case 202:
			un_link_commands();
			exit(0);
		case 'h':
			usage(program_name);
			help();
			exit(-1);
		default:
			usage(program_name);
			exit(-1);
		}
	if(autowin != NULL) {
		auto_win_flag=atoi(autowin);
	}
	//cartel();
	cam = open_camera();

	if( cam > 0 ) {
		ret = 0;
		set_location();		
		command = find_command(program_name);
		if (!command && argv[optind]) {
			command = find_command(argv[optind]);
			if (command) optind++;
		}
		if (command) {
			readline_flag = 0;
			readline_history_flag = 0;
			interactive_flag = 0;
			line[0] = 0;
			if (optind < argc) {
				while (optind < argc) {
					strcat(line, argv[optind++]);
					if (optind < argc)
						strcat(line, " ");
				}
			}
			//printf("CMD: [%s %s]\n", command->name, line);
			//fflush(stdout);
			ret = ((*(command->func)) (line));
		} else {
			cartel();
			caxpa_open();
			command_main();
		}
		close_camera(cam);
		caxpa_close();
	}else{
		PVERB(PERR," stari: can not connect to cc\n");
	}

        return ret;
}

